# Copyright 2025 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from requests import Response
from pandas import DataFrame, Series
from typing import Union, Optional, List
from spgci.api_client import get_data, Paginator
from spgci.utilities import list_to_filter
from enum import Enum
import warnings


class GlobalOilDemand:
    """
    Global Oil Demand

    .. deprecated:: 0.0.57
       Use `OilNGLAnalytics` instead.
    """

    def __init__(self):
        warnings.warn(
            "GlobalOilDemand is deprecated. Use OilNGLAnalytics instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    _path = "mgodf/v1/"

    class RefTypes(Enum):
        """Global Oil Demand Reference Data Types"""

        Countries = "countries"
        DataQuality = "dataquality"
        Products = "products"
        Regions = "regions"
        Scenarios = "scenario-list"

    @staticmethod
    def _paginate(resp: Response) -> Paginator:
        j = resp.json()
        count = j["metadata"]["count"]
        size = j["metadata"]["pageSize"]

        remainder = count % size
        quotient = count // size
        total_pages = quotient + (1 if remainder > 0 else 0)

        if total_pages <= 1:
            return Paginator(False, "page", 1)

        return Paginator(True, "page", total_pages=total_pages)

    def get_demand(
        self,
        *,
        year: Optional[Union[int, list[int], "Series[int]"]] = None,
        year_gt: Optional[int] = None,
        year_gte: Optional[int] = None,
        year_lt: Optional[int] = None,
        year_lte: Optional[int] = None,
        month: Optional[Union[int, list[int]]] = None,
        product: Optional[Union[str, list[str], "Series[str]"]] = None,
        region: Optional[Union[str, list[str], "Series[str]"]] = None,
        country: Optional[Union[str, list[str], "Series[str]"]] = None,
        units: Optional[Union[str, list[str], "Series[str]"]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        .. deprecated:: 0.0.57
            Use `OilNGLAnalytics` instead.

        Fetch latest Global Oil Demand forecast by month.

        Parameters
        ----------
        year : Optional[Union[int, list[int], Series[int]]], optional
            filter by year, by default None
        year_gt : Optional[int], optional
            filter by ``year > x``, by default None
        year_gte : Optional[int], optional
            filter by ``year >= x``, by default None
        year_lt : Optional[int], optional
            filter by ``year < x``, by default None
        year_lte : Optional[int], optional
            filter by ``year <= x``, by default None
        product : Optional[Union[str, list[str], Series[str]]], optional
            filter by productName, by default None
        region : Optional[Union[str, list[str], Series[str]]], optional
            filter by regionName, by default None
        country : Optional[Union[str, list[str], Series[str]]], optional
            filter by countryName, by default None
        units : Optional[Union[str, list[str], Series[str]]], optional
            filter by units, by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
              DataFrame
                  DataFrame of the ``response.json()``
              Response
                  Raw ``requests.Response`` object
        Examples
        --------
        **Simple Example**
        >>> ci.GlobalOilDemand().get_demand(year=2022)

        **Using Lists**
        >>> ci.GlobalOilDemand().get_demand(country="Cambodia", product=["Naphtha", "Ethane"])

        **Compose with Ref Data**
        >>> products = ci.GlobalOilDemand().get_reference_data(type=ci.GlobalOilDemand.RefTypes.Products)
        >>> ci.GlobalOilDemand().get_demand(product=products["productName"][:2], year=2023)
        """
        endpoint_path = "demand"

        filter_param: List[str] = []

        filter_param.append(list_to_filter("year", year))
        filter_param.append(list_to_filter("productName", product))
        filter_param.append(list_to_filter("month", month))
        filter_param.append(list_to_filter("regionName", region))
        filter_param.append(list_to_filter("countryName", country))
        filter_param.append(list_to_filter("units", units))

        if year_gt:
            filter_param.append(f"year > {year_gt}")
        if year_gte:
            filter_param.append(f"year >= {year_gte}")
        if year_lt:
            filter_param.append(f"year < {year_lt}")
        if year_lte:
            filter_param.append(f"year <= {year_lte}")

        filter_param = [fp for fp in filter_param if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_param)
        elif len(filter_param) > 0:
            filter_exp = " AND ".join(filter_param) + " AND (" + filter_exp + ")"

        params = {
            "pageSize": page_size,
            "filter": filter_exp,
            "page": page,
        }
        return get_data(
            path=f"{self._path}{endpoint_path}",
            params=params,
            paginate=paginate,
            paginate_fn=self._paginate,
            raw=raw,
        )

    def get_demand_archive(
        self,
        scenario_id: int,
        *,
        year: Optional[Union[int, list[int], "Series[int]"]] = None,
        year_gt: Optional[int] = None,
        year_gte: Optional[int] = None,
        year_lt: Optional[int] = None,
        year_lte: Optional[int] = None,
        month: Optional[Union[int, list[int]]] = None,
        product: Optional[Union[str, list[str], "Series[str]"]] = None,
        region: Optional[Union[str, list[str], "Series[str]"]] = None,
        country: Optional[Union[str, list[str], "Series[str]"]] = None,
        units: Optional[Union[str, list[str], "Series[str]"]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        .. deprecated:: 0.0.57
            Use `OilNGLAnalytics` instead.

        Fetch archive Global Oil Demand forecast by month.

        Parameters
        ----------
        scenario_id : int
            filter by scenarioId
        year : Optional[Union[int, list[int], Series[int]]], optional
            filter by year, by default None
        year_gt : Optional[int], optional
            filter by ``year > x``, by default None
        year_gte : Optional[int], optional
            filter by ``year >= x``, by default None
        year_lt : Optional[int], optional
            filter by ``year < x``, by default None
        year_lte : Optional[int], optional
            filter by ``year <= x``, by default None
        product : Optional[Union[str, list[str], Series[str]]], optional
            filter by productName, by default None
        region : Optional[Union[str, list[str], Series[str]]], optional
            filter by regionName, by default None
        country : Optional[Union[str, list[str], Series[str]]], optional
            filter by countryName, by default None
        units : Optional[Union[str, list[str], Series[str]]], optional
            filter by units, by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
              DataFrame
                  DataFrame of the ``response.json()``
              Response
                  Raw ``requests.Response`` object
        Examples
        --------
        **Simple Example**
        >>> ci.GlobalOilDemand().get_demand_archive(scenario_id=150, country="Norway")

        **Using Lists**
        >>> ci.GlobalOilDemand().get_demand_archive(scenario_id=150, country="Cambodia", product=["Naphtha", "Ethane"])

        **Compose with Ref Data**
        >>> scenarios = ci.GlobalOilDemand().get_reference_data(type=ci.GlobalOilDemand.RefTypes.Scenarios)
        >>> ci.GlobalOilDemand().get_demand_archive(scenario_id=scenarios['scenarioId'].iloc[-1], year=2022)
        """
        endpoint_path = "demand-archive"

        filter_param: List[str] = []

        filter_param.append(list_to_filter("year", year))
        filter_param.append(list_to_filter("productName", product))
        filter_param.append(list_to_filter("month", month))
        filter_param.append(list_to_filter("regionName", region))
        filter_param.append(list_to_filter("countryName", country))
        filter_param.append(list_to_filter("units", units))

        if year_gt:
            filter_param.append(f"year > {year_gt}")
        if year_gte:
            filter_param.append(f"year >= {year_gte}")
        if year_lt:
            filter_param.append(f"year < {year_lt}")
        if year_lte:
            filter_param.append(f"year <= {year_lte}")

        filter_param = [fp for fp in filter_param if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_param)
        elif len(filter_param) > 0:
            filter_exp = " AND ".join(filter_param) + " AND (" + filter_exp + ")"

        params = {
            "scenarioId": scenario_id,
            "pageSize": page_size,
            "filter": filter_exp,
            "page": page,
        }
        return get_data(
            path=f"{self._path}{endpoint_path}",
            params=params,
            paginate=paginate,
            paginate_fn=self._paginate,
            raw=raw,
        )

    def get_reference_data(
        self, type: RefTypes, raw: bool = False
    ) -> Union[Response, DataFrame]:
        """
        .. deprecated:: 0.0.57
            Use `OilNGLAnalytics` instead.

        Fetch reference data for the Global Oil Demand dataset.

        Parameters
        ----------
        type : RefTypes
            filter by type
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
                DataFrame of the ``response.json()``
            Response
                Raw ``requests.Response`` object
        Examples
        --------
        **Simple**
        >>> ci.GlobalOilDemand().get_reference_data(type=ci.GlobalOilDeman.RefTypes.Products)
        """
        endpoint_path = type.value

        params = {"pageSize": 1000}

        return get_data(
            path=f"{self._path}{endpoint_path}",
            params=params,
            paginate=True,
            paginate_fn=self._paginate,
            raw=raw,
        )



# Copyright 2025 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from requests import Response
from pandas import DataFrame, Series, to_datetime  # type: ignore
from typing import Union, Optional, List
from spgci.api_client import get_data, Paginator
from spgci.utilities import list_to_filter
from enum import Enum
from datetime import date
from typing_extensions import Literal
import pandas as pd
from packaging.version import parse


class Arbflow:
    """
    Refining Margins & Crude Arbitrage

    Includes
    --------
    ``RefTypes`` to use with the ``get_reference_data`` method.
    ``get_reference_data()`` to get the list of locations, crudes, configurations, etc.. for the Refining Margins & Crude Arbitrage dataset.
    ``get_margins_catalog()`` to get the refining margins catalog which shows the unique combination of factors
                            (the crude, location, refinery configuration and freight) that describe the margins.
    ``get_margins_data()`` to get the refining margins data.
    ``get_arbitrage()`` get the crude arbitrage data.

    """

    _path = "arbflow/data/v1/"

    class RefTypes(Enum):
        """Refining Margins & Crude Arbitrage Reference Data Types"""

        Configurations = "configurations"
        Crudes = "crudes"
        Frequencies = "frequencies"
        Locations = "locations"
        Margins = "margins"

    @staticmethod
    def _paginate(resp: Response) -> Paginator:
        j = resp.json()
        count = j["metadata"]["count"]
        size = j["metadata"]["pageSize"]

        remainder = count % size
        quotient = count // size
        total_pages = quotient + (1 if remainder > 0 else 0)

        if total_pages <= 1:
            return Paginator(False, "page", 1)

        return Paginator(True, "page", total_pages=total_pages)

    @staticmethod
    def _convert_to_df(resp: Response) -> DataFrame:
        j = resp.json()
        df = DataFrame(j["results"])

        date_columns = ["modifiedDate", "marginDate", "baseMarginDate"]

        for column in date_columns:
            if column in df.columns:
                if parse(pd.__version__) >= parse("2"):
                    df[column] = pd.to_datetime(
                        df[column], format="ISO8601", errors="coerce"
                    )
                else:
                    df[column] = pd.to_datetime(df[column], errors="coerce")  # type: ignore

        return df

    def get_margins_catalog(
        self,
        *,
        margin_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        margin_name: Optional[Union[str, list[str], "Series[str]"]] = None,
        crude_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        crude_name: Optional[Union[str, list[str], "Series[str]"]] = None,
        crude_symbol: Optional[Union[str, list[str], "Series[str]"]] = None,
        location_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        location_name: Optional[Union[str, list[str], "Series[str]"]] = None,
        configuration_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        configuration_name: Optional[Union[str, list[str], "Series[str]"]] = None,
        preferred_freight: Optional[bool] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Get the refining margins catalog which shows the unique combination of factors (the crude, location,
        refinery configuration and freight) that describe the margins.

        Parameters
        ----------
        margin_id   :    Optional[Union[int, list[int], "Series[int]"]] = None,
            filter by marginId, by default None
        margin_name :    Optional[Union[str, list[str], "Series[str]"]] = None,
            filter by marginName, by default None
        crude_id    :    Optional[Union[int, list[int], "Series[int]"]] = None,
            filter by crudeId, by default None
        crude_name  :    Optional[Union[str, list[str], "Series[str]"]] = None,
            filter by crudeName, by default None
        crude_symbol    :    Optional[Union[str, list[str], "Series[str]"]] = None,
            filter by crudeSymbol, by default None
        location_id     :    Optional[Union[int, list[int], "Series[int]"]] = None,
            filter by locationId, by default None
        location_name   :   Optional[Union[str, list[str], "Series[str]"]] = None,
            filter by locationName, by default None
        configuration_id   :    Optional[Union[int, list[int], "Series[int]"]] = None,
            filter by configurationId, by default None
        configuration_name  :   Optional[Union[str, list[str], "Series[str]"]] = None,
            filter by configurationName, by default None
        preferred_freight   :   Optional[bool] = None,
            filter by preferredFreight, by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
              DataFrame
                  DataFrame of the ``response.json()``
              Response
                  Raw ``requests.Response`` object
        Examples
        --------
        **Get Margins Catalog**
        >>> ci.Arbflow().get_margins_catalog(location_id = 34)

        **Get Margins Catalog for multiple locations **
        >>> ci.Arbflow().get_margins_catalog(location_id = [31,34])

        **Get Margins Catalog Data Based On Location Id And Crude Symbol**
        >>> ci.Arbflow().get_margins_catalog(location_id = 34, crude_symbol="AAQZB00")

        """

        endpoint_path = "margins-catalog"

        filter_param: List[str] = []

        filter_param.append(list_to_filter("marginId", margin_id))
        filter_param.append(list_to_filter("marginName", margin_name))
        filter_param.append(list_to_filter("crudeId", crude_id))
        filter_param.append(list_to_filter("crudeName", crude_name))
        filter_param.append(list_to_filter("crudeSymbol", crude_symbol))
        filter_param.append(list_to_filter("locationId", location_id))
        filter_param.append(list_to_filter("locationName", location_name))
        filter_param.append(list_to_filter("configurationId", configuration_id))
        filter_param.append(list_to_filter("configurationName", configuration_name))
        filter_param.append(list_to_filter("preferredFreight", preferred_freight))

        filter_param = [fp for fp in filter_param if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_param)
        elif len(filter_param) > 0:
            filter_exp = " AND ".join(filter_param) + " AND (" + filter_exp + ")"

        params = {
            "pageSize": page_size,
            "filter": filter_exp,
            "page": page,
        }
        return get_data(
            path=f"{self._path}{endpoint_path}",
            params=params,
            paginate=paginate,
            paginate_fn=self._paginate,
            df_fn=self._convert_to_df,
            raw=raw,
        )

    def get_margins_data(
        self,
        frequency_id: Literal[1, 2, 3] = 1,
        *,
        margin_id: Optional[Union[int, list[int], "Series[int]"]] = None,
        margin_date: Optional[date] = None,
        margin_date_gt: Optional[date] = None,
        margin_date_gte: Optional[date] = None,
        margin_date_lt: Optional[date] = None,
        margin_date_lte: Optional[date] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Get the refining margins data.

        Parameters
        ----------
        margin_id   :    Optional[Union[int, list[int], "Series[int]"]] = None,
            filter by marginId, by default None
        frequency_id    :    Literal[1,2,3] = 1,
            filter by frequencyId - 1 = Daily, 2 = Monthly, 3 = Annual, by default 1
        margin_date: Optional[date] = None,
            filter by marginDate, by default None
        margin_date_lt : Optional[date], optional
            filter by ``marginDate < x``, by default None
        margin_date_lte : Optional[date], optional
            filter by ``marginDate <= x``, by default None
        margin_date_gt : Optional[date], optional
            filter by ``marginDate > x``, by default None
        margin_date_gte : Optional[date], optional
            filter by ``marginDate >= x``, by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
              DataFrame
                  DataFrame of the ``response.json()``
              Response
                  Raw ``requests.Response`` object
        Examples
        --------
        **Get Margins Data**
        >>> ci.Arbflow().get_margins_data(frequency_id = 1)

        **Get Margins Data Based On Margin Id And Date**
        >>> ci.Arbflow().get_margins_data(margin_id=229, margin_date='2023-08-16')

        """

        endpoint_path = "margins-data"

        filter_param: List[str] = []

        filter_param.append(list_to_filter("marginId", margin_id))
        if margin_date:
            filter_param.append(f'marginDate: "{margin_date}"')
        if margin_date_gt:
            filter_param.append(f'marginDate > "{margin_date_gt}"')
        if margin_date_gte:
            filter_param.append(f'marginDate >= "{margin_date_gte}"')
        if margin_date_lt:
            filter_param.append(f'marginDate < "{margin_date_lt}"')
        if margin_date_lte:
            filter_param.append(f'marginDate <= "{margin_date_lte}"')

        filter_param = [fp for fp in filter_param if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_param)
        elif len(filter_param) > 0:
            filter_exp = " AND ".join(filter_param) + " AND (" + filter_exp + ")"

        params = {
            "frequencyId": frequency_id,
            "pageSize": page_size,
            "filter": filter_exp,
            "page": page,
        }
        return get_data(
            path=f"{self._path}{endpoint_path}",
            params=params,
            paginate=paginate,
            paginate_fn=self._paginate,
            df_fn=self._convert_to_df,
            raw=raw,
        )

    def get_arbitrage(
        self,
        base_margin_id: int,
        margin_id: Union[list[int], "Series[int]", int],
        frequency_id: Literal[1, 2, 3] = 1,
        *,
        base_margin_date: Optional[date] = None,
        start_date: Optional[date] = None,
        end_date: Optional[date] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Get the crude arbitrage data..

        Parameters
        ----------
        margin_id   :  Union[list[int], "Series[int]", int],
            filter by comparisonMarginId, maximum of 5 comparison margin IDs. (required field)
        base_margin_id: int,
            filter by baseMarginId (required field)
        frequency_id    :    Literal[1,2,3] = 1,
            filter by frequencyId - 1 = Daily, 2 = Monthly, 3 = Annual, by default 1 (required field)
        base_margin_date: Optional[datetime] = None,
            filter by marginDate, by default None
        start_date: Optional[datetime] = None,
            filter by startDate, by default None
        end_date: Optional[datetime] = None,
            filter by endDate, by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
              DataFrame
                  DataFrame of the ``response.json()``
              Response
                  Raw ``requests.Response`` object
        Examples
        --------
        **Get Arbitrage Data**
        >>> ci.Arbflow().get_arbitrage(margin_id=229, base_margin_id=330, frequency_id=1)

        **Get Arbitrage Data Based On Multiple Comparison Margin Id's **
        >>> ci.Arbflow().get_arbitrage(margin_id=[220,330], base_margin_id =330, frequency_id=1)

        **Using Date Range**
        >>> ci.Arbflow().get_arbitrage(frequency_id=1, margin_id=261, base_margin_id=1380, start_date="2024-01-01", end_date="2024-01-31")
        """

        endpoint_path = "arbitrage"

        filter_param: List[str] = []

        filter_param.append(list_to_filter("baseMarginDate", base_margin_date))

        filter_param = [fp for fp in filter_param if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_param)
        elif len(filter_param) > 0:
            filter_exp = " AND ".join(filter_param) + " AND (" + filter_exp + ")"

        if isinstance(margin_id, int):
            cmp_margin_id = margin_id
        elif isinstance(margin_id, list):
            cmp_margin_id = ",".join(map(str, margin_id))
        else:
            # redundant / maybe user will pass a string.
            cmp_margin_id = margin_id

        params = {
            "ComparisonMarginIds": cmp_margin_id,
            "baseMarginId": base_margin_id,
            "frequencyId": frequency_id,
            "pageSize": page_size,
            "filter": filter_exp,
            "page": page,
        }

        if start_date is not None:
            params["startDate"] = start_date
        if end_date is not None:
            params["endDate"] = end_date

        return get_data(
            path=f"{self._path}{endpoint_path}",
            params=params,
            paginate=paginate,
            paginate_fn=self._paginate,
            df_fn=self._convert_to_df,
            raw=raw,
        )

    def get_reference_data(
        self, type: RefTypes, raw: bool = False
    ) -> Union[Response, DataFrame]:
        """
        Fetch reference data for the Refining Margins & Crude Arbitrage dataset.

        Parameters
        ----------
        type : RefTypes
            filter by type
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
                DataFrame of the ``response.json()``
            Response
                Raw ``requests.Response`` object
        Examples
        --------
        **Simple**
        >>> ci.Arbflow().get_reference_data(type=ci.Arbflow.RefTypes.Crudes)
        >>> ci.Arbflow().get_reference_data(type=ci.Arbflow.RefTypes.Margins)
        >>> ci.Arbflow().get_reference_data(type=ci.Arbflow.RefTypes.Locations)
        """
        endpoint_path = type.value

        params = {"pageSize": 1000}

        return get_data(
            path=f"{self._path}{endpoint_path}",
            params=params,
            paginate=True,
            paginate_fn=self._paginate,
            df_fn=self._convert_to_df,
            raw=raw,
        )



# Copyright 2025 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module to handle api request"""
import requests
import spgci.config
from spgci.auth import get_token
from typing import Callable, Dict, Any, NamedTuple, Union
from pandas import DataFrame
import pandas as pd
from tqdm import tqdm
import warnings
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed
from spgci.exceptions import AuthError, PerSecondLimitError, DailyLimitError
from time import sleep
from urllib.parse import urlparse, parse_qsl, urlencode, quote

_auth_retry = retry(
    retry=retry_if_exception_type(AuthError), reraise=True, stop=stop_after_attempt(2)
)
_throttle_retry = retry(
    retry=retry_if_exception_type(PerSecondLimitError), reraise=True, wait=wait_fixed(1)
)


class Paginator(NamedTuple):
    has_more_pages: bool
    key: str
    total_pages: int
    next_link: str = ""
    pg_type: str = ""


def _to_df(resp: requests.Response) -> DataFrame:
    j = resp.json()
    return DataFrame(j["results"])


def _nop_paginate(resp: requests.Response) -> Paginator:
    return Paginator(False, "page", 1)


def _paginate(resp: requests.Response) -> Paginator:
    j = resp.json()
    total_pages = j["metadata"]["totalPages"]

    if total_pages <= 1:
        return Paginator(False, "page", 1)

    return Paginator(True, "page", total_pages=total_pages)


_session = requests.Session()


@_auth_retry
@_throttle_retry
def _get(
    url: str,
    params: Dict[Any, Any],
    session: requests.Session,
    # token_fn: Callable[[str, str, str, str], str] = get_token,
) -> requests.Response:
    headers = {
        "User-Agent": f"spgci-py/{spgci.config.version}",
        # "Authorization": f"Bearer {token}",
    }

    if spgci.config.get_token() is not None:
        token = spgci.config.get_token()
    else:
        token = spgci.auth.get_token(
            spgci.config.username,
            spgci.config.password,
            spgci.config.base_url,
        )

    headers["Authorization"] = f"Bearer {token}"

    # should remove at some point..
    sleep(spgci.config.sleep_time)

    response: requests.Response = session.get(
        url=url,
        params=params,
        headers=headers,
        verify=spgci.config.verify_ssl,
        proxies=spgci.config.proxies,
        auth=spgci.config.auth,
    )

    # clear token cache and retry once if its a 401/403. shouldn't be hit unless token is expired..
    if response.status_code in [401, 403]:
        get_token.cache_clear()
        raise AuthError("Invalid Username, Password or Appkey")

    # if 429 check if more requests can be made today.
    if response.status_code == 429:
        rl = int(response.headers.get("x-ratelimit-remaining-day", 0))
        if rl > 0:
            raise PerSecondLimitError("Per Second Rate Limit Reached")
        else:
            raise DailyLimitError("Daily Rate Limit Reached")

    if response.status_code != 200:
        print(response.text)
        response.raise_for_status()

    return response


def get_data(
    path: str,
    params: Dict[Any, Any],
    df_fn: Callable[[requests.Response], DataFrame] = _to_df,
    paginate_fn: Callable[[requests.Response], Paginator] = _paginate,
    raw: bool = False,
    paginate: bool = False,
) -> Union[DataFrame, requests.Response]:
    url = f"{spgci.config.base_url}/{path}"
    response = _get(url, params=params, session=_session)

    if raw:
        if paginate:
            warnings.warn(
                f"\nCannot set `paginate=True` along with `raw=True`. Returning only the page requested."
            )
        return response

    df: DataFrame = df_fn(response)
    pagination = paginate_fn(response)

    if not pagination.has_more_pages:
        return df

    if not paginate:
        warnings.warn(
            f"\nFetched page [1] of [{pagination.total_pages}]. set `paginate=True` to fetch all pages."
        )
        return df

    tp = pagination.total_pages
    if tp > 10:
        if spgci.config.is_agent:
            raise ValueError(
                f"Agent mode is enabled. Cannot paginate when total pages ({tp}) > 10. "
                f"Set `paginate=False` to fetch only the first page."
            )
        warnings.warn(
            f"\nWith `paginate=True` this will fetch {tp} pages. Set `paginate=False` to disable."
        )
    for i in tqdm(
        range(2, tp + 1),
        desc="Fetching...",
        initial=1,
        total=tp,
    ):
        # special handling for oData. Should refactor this at some point
        # also the pageSize param is hardcoded..
        if pagination.pg_type == "odata":
            parsed = urlparse(url)
            qs = dict(parse_qsl(parsed.query))
            qs[pagination.key] = str((i - 1) * int(qs["pageSize"]))
            parsed = parsed._replace(query=urlencode(qs, quote_via=quote))
            resp = _get(url=parsed.geturl(), params={}, session=_session)
        else:
            params[pagination.key] = i
            resp = _get(
                url,
                params=params,
                session=_session,
            )

        new_df = df_fn(resp)
        df: DataFrame = pd.concat(objs=[df, new_df], ignore_index=True)  # type: ignore

    return df



# Copyright 2025 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from .api_client import get_data, Paginator
from .utilities import list_to_filter
from typing import Union, Optional
from typing_extensions import Literal
from pandas import DataFrame, Series, to_datetime  # type: ignore
from requests import Response
from enum import Enum
from datetime import date


class EnergyPriceForecast:
    """
    The Energy Price Forecast API provides a comprehensive view of S&P Global Platts latest energy price forecasts and historical monthly/yearly averages.
    Short-term forecasts are available up to 18-months out on a monthly granularity and long-term forecasts are available on a yearly granularity up to 30-years out.
    The Energy Price Forecast API also gives users access to historical forecasts to easily compare how Platts' outlook has evolved over time.

    """

    _path = "/energy-price-forecast/v1/"

    @staticmethod
    def _paginate(resp: Response) -> Paginator:
        j = resp.json()
        count = j["metadata"]["count"]
        size = j["metadata"]["pageSize"]

        remainder = count % size
        quotient = count // size

        total_pages = quotient + (1 if remainder > 0 else 0)

        if total_pages <= 1:
            return Paginator(False, "page", 1)

        return Paginator(True, "page", total_pages=total_pages)

    @staticmethod
    def _to_df(resp: Response) -> DataFrame:
        j = resp.json()
        df = DataFrame(j["results"])

        if len(df) > 0:
            df["modifiedDate"] = to_datetime(df["modifiedDate"])

        return df

    class RefTypes(Enum):
        """Reference Types to use with the `get_reference_date()` method"""

        Commodities = "commodities"
        Categories = "categories"
        Currencies = "currencies"
        DeliveryRegions = "delivery-regions"
        Groups = "groups"
        Prices = "prices"
        # LTArchiveDates = "long-term-archive-dates" handle these in a separate method.
        # STArchiveDates = "short-term-archive-dates"
        Sectors = "sectors"
        SectorGroups = "sector-groups"
        Units = "units"
        LTYears = "years-long-term"
        STYears = "years-short-term"
        # LTYearsCategory = "years-lt-category" unsupported for now.
        # STYearsCategory = "years-st-category"

    def get_prices_shortterm(
        self,
        *,
        year: Optional[Union[list[int], "Series[int]", int]] = None,
        year_gt: Optional[int] = None,
        year_gte: Optional[int] = None,
        year_lt: Optional[int] = None,
        year_lte: Optional[int] = None,
        month: Optional[Union[list[int], "Series[int]", int]] = None,
        symbol: Optional[Union[list[str], "Series[str]", str]] = None,
        units: Optional[Union[list[str], "Series[str]", str]] = None,
        currency: Optional[Union[list[str], "Series[str]", str]] = None,
        category: Optional[Union[list[str], "Series[str]", str]] = None,
        group: Optional[Union[list[str], "Series[str]", str]] = None,
        sector: Optional[Union[list[str], "Series[str]", str]] = None,
        commodity: Optional[Union[list[str], "Series[str]", str]] = None,
        delivery_region: Optional[Union[list[str], "Series[str]", str]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[DataFrame, Response]:
        """
        Latest short-term price forecasts provided on a monthly granularity.

        Parameters
        ----------
        year : Optional[Union[int, list[int], Series[int]]], optional
            filter by ``year = x``, by default None
        year_gt : Optional[int], optional
            filter by ``year > x``, by default None
        year_gte : Optional[int], optional
            filter by ``year >= x``, by default None
        year_lt : Optional[int], optional
            filter by ``year < x``, by default None
        year_lte : Optional[int], optional
            filter by ``year <= x``, by default None
        month : Optional[Union[list[int], Series[int], int]], optional
            filter by month, by default None
        symbol : Optional[Union[list[str], Series[str], str]], optional
            filter by priceSymbol, by default None
        units : Optional[Union[list[str], Series[str], str]], optional
            filter by unitName, by default None
        currency : Optional[Union[list[str], Series[str], str]], optional
            filter by currencySymbol, by default None
        category : Optional[Union[list[str], Series[str], str]], optional
            filter by categoryName, by default None
        group : Optional[Union[list[str], Series[str], str]], optional
            filter by groupName, by default None
        sector : Optional[Union[list[str], Series[str], str]], optional
            filter by sectorName, by default None
        commodity : Optional[Union[list[str], Series[str], str]], optional
            filter by commodityName, by default None
        delivery_region : Optional[Union[list[str], Series[str], str]], optional
            filter by deliveryRegionName, by default None
        filter_exp : Optional[str], optional
            pass-thru ``$filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to select a certain page number, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object

        Examples
        --------
        **Simple**
        >>> ci.EnergyPriceForecast().get_prices_shortterm(symbol="PCAAS00", group="Crude")

        **Using Lists**
        >>> ci.EnergyPriceForecast().get_prices_shortterm(month=[10, 11, 12], group="Crude", delivery_region="United States")

        **Using Series**
        >>> groups = ci.EnergyPriceForecast().get_reference_data(type=ci.EnergyPriceForecast.RefTypes.SectorGroups)
        >>> ci.EnergyPriceForecast().get_prices_shortterm(group=groups['groupName'].tail(2))
        """
        path = "prices-short-term"
        filter_params: list[str] = []

        if year_gt != None:
            filter_params.append(f"year > {year_gt}")
        if year_gte != None:
            filter_params.append(f"year >= {year_gte}")
        if year_lt != None:
            filter_params.append(f"year < {year_lt}")
        if year_lte != None:
            filter_params.append(f"year <= {year_lte}")

        filter_params.append(list_to_filter("year", year))
        filter_params.append(list_to_filter("priceSymbol", symbol))
        filter_params.append(list_to_filter("month", month))
        filter_params.append(list_to_filter("unitName", units))
        filter_params.append(list_to_filter("categoryName", category))
        filter_params.append(list_to_filter("groupName", group))
        filter_params.append(list_to_filter("currencySymbol", currency))
        filter_params.append(list_to_filter("commodityName", commodity))
        filter_params.append(list_to_filter("deliveryRegionName", delivery_region))
        filter_params.append(list_to_filter("sectorName", sector))

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {"filter": filter_exp, "page": page, "pageSize": page_size}
        return get_data(
            path=f"{self._path}{path}",
            params=params,
            paginate=paginate,
            raw=raw,
            paginate_fn=self._paginate,
            df_fn=self._to_df,
        )

    def get_prices_longterm(
        self,
        *,
        year: Optional[Union[list[int], "Series[int]", int]] = None,
        year_gt: Optional[int] = None,
        year_gte: Optional[int] = None,
        year_lt: Optional[int] = None,
        year_lte: Optional[int] = None,
        symbol: Optional[Union[list[str], "Series[str]", str]] = None,
        units: Optional[Union[list[str], "Series[str]", str]] = None,
        currency: Optional[Union[list[str], "Series[str]", str]] = None,
        category: Optional[Union[list[str], "Series[str]", str]] = None,
        group: Optional[Union[list[str], "Series[str]", str]] = None,
        sector: Optional[Union[list[str], "Series[str]", str]] = None,
        commodity: Optional[Union[list[str], "Series[str]", str]] = None,
        delivery_region: Optional[Union[list[str], "Series[str]", str]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[DataFrame, Response]:
        """
        Latest long-term price forecasts provided on a yearly granularity.

        Parameters
        ----------
        year : Optional[Union[int, list[int], Series[int]]], optional
            filter by ``year = x``, by default None
        year_gt : Optional[int], optional
            filter by ``year > x``, by default None
        year_gte : Optional[int], optional
            filter by ``year >= x``, by default None
        year_lt : Optional[int], optional
            filter by ``year < x``, by default None
        year_lte : Optional[int], optional
            filter by ``year <= x``, by default None
        symbol : Optional[Union[list[str], Series[str], str]], optional
            filter by priceSymbol, by default None
        units : Optional[Union[list[str], Series[str], str]], optional
            filter by unitName, by default None
        currency : Optional[Union[list[str], Series[str], str]], optional
            filter by currencySymbol, by default None
        category : Optional[Union[list[str], Series[str], str]], optional
            filter by categoryName, by default None
        group : Optional[Union[list[str], Series[str], str]], optional
            filter by groupName, by default None
        sector : Optional[Union[list[str], Series[str], str]], optional
            filter by sectorName, by default None
        commodity : Optional[Union[list[str], Series[str], str]], optional
            filter by commodityName, by default None
        delivery_region : Optional[Union[list[str], Series[str], str]], optional
            filter by deliveryRegionName, by default None
        filter_exp : Optional[str], optional
            pass-thru ``$filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to select a certain page number, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object

        Examples
        --------
        **Simple**
        >>> ci.EnergyPriceForecast().get_prices_longterm(year=2040)

        **Using Lists**
        >>> ci.EnergyPriceForecast().get_prices_longterm(group=["Crude", "Common Crude Spreads"], delivery_region="United States")

        **Using Series**
        >>> years = ci.EnergyPriceForecast().get_reference_data(type=ci.EnergyPriceForecast.RefTypes.LTYears)
        >>> ci.EnergyPriceForecast().get_prices_longterm(year=years['year'][:2])
        """
        path = "prices-long-term"
        filter_params: list[str] = []

        if year_gt != None:
            filter_params.append(f"year > {year_gt}")
        if year_gte != None:
            filter_params.append(f"year >= {year_gte}")
        if year_lt != None:
            filter_params.append(f"year < {year_lt}")
        if year_lte != None:
            filter_params.append(f"year <= {year_lte}")

        filter_params.append(list_to_filter("year", year))
        filter_params.append(list_to_filter("priceSymbol", symbol))
        filter_params.append(list_to_filter("unitName", units))
        filter_params.append(list_to_filter("categoryName", category))
        filter_params.append(list_to_filter("groupName", group))
        filter_params.append(list_to_filter("currencySymbol", currency))
        filter_params.append(list_to_filter("commodityName", commodity))
        filter_params.append(list_to_filter("deliveryRegionName", delivery_region))
        filter_params.append(list_to_filter("sectorName", sector))

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {"filter": filter_exp, "page": page, "pageSize": page_size}
        return get_data(
            path=f"{self._path}{path}",
            params=params,
            paginate=paginate,
            raw=raw,
            paginate_fn=self._paginate,
            df_fn=self._to_df,
        )

    def get_prices_shortterm_archive(
        self,
        *,
        modified_date: date,
        category_id: int,
        month: Optional[Union[list[int], "Series[int]", int]] = None,
        year: Optional[Union[list[int], "Series[int]", int]] = None,
        year_gt: Optional[int] = None,
        year_gte: Optional[int] = None,
        year_lt: Optional[int] = None,
        year_lte: Optional[int] = None,
        symbol: Optional[Union[list[str], "Series[str]", str]] = None,
        units: Optional[Union[list[str], "Series[str]", str]] = None,
        currency: Optional[Union[list[str], "Series[str]", str]] = None,
        category: Optional[Union[list[str], "Series[str]", str]] = None,
        group: Optional[Union[list[str], "Series[str]", str]] = None,
        sector: Optional[Union[list[str], "Series[str]", str]] = None,
        commodity: Optional[Union[list[str], "Series[str]", str]] = None,
        delivery_region: Optional[Union[list[str], "Series[str]", str]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[DataFrame, Response]:
        """
        Archived short-term price forecasts.

        see `get_archive_dates()` for applicable modified dates/categories.

        Parameters
        ----------
        modified_date : date
            filter on ModifiedDate to see the nearest prior forecast
        category_id : int
            filter on PriceCategoryCode
        year : Optional[Union[int, list[int], Series[int]]], optional
            filter by ``year = x``, by default None
        year_gt : Optional[int], optional
            filter by ``year > x``, by default None
        year_gte : Optional[int], optional
            filter by ``year >= x``, by default None
        year_lt : Optional[int], optional
            filter by ``year < x``, by default None
        year_lte : Optional[int], optional
            filter by ``year <= x``, by default None
        month : Optional[Union[list[int], Series[int], int]], optional
            filter by month, by default None
        symbol : Optional[Union[list[str], Series[str], str]], optional
            filter by priceSymbol, by default None
        units : Optional[Union[list[str], Series[str], str]], optional
            filter by unitName, by default None
        currency : Optional[Union[list[str], Series[str], str]], optional
            filter by currencySymbol, by default None
        category : Optional[Union[list[str], Series[str], str]], optional
            filter by categoryName, by default None
        group : Optional[Union[list[str], Series[str], str]], optional
            filter by groupName, by default None
        sector : Optional[Union[list[str], Series[str], str]], optional
            filter by sectorName, by default None
        commodity : Optional[Union[list[str], Series[str], str]], optional
            filter by commodityName, by default None
        delivery_region : Optional[Union[list[str], Series[str], str]], optional
            filter by deliveryRegionName, by default None
        filter_exp : Optional[str], optional
            pass-thru ``$filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to select a certain page number, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object

        Examples
        --------
        **Simple**
        >>> from datetime import date
        >>> ci.EnergyPriceForecast().get_prices_shortterm_archive(modified_date=date(2020, 11, 1), category_id=1, year=2022)

        **Using Lists**
        >>> from datetime import date
        >>> ci.EnergyPriceForecast().get_prices_shortterm_archive(modified_date=date(2020, 11, 1) category_id=1, year=2022, group=["Refinery Margins", "U.S. West Coast"])

        **Using Series**
        >>> cat_dates = ci.EnergyPriceForecast().get_archive_dates(term_type="short", category="Oil")
        >>> cat_dates = cat_dates.iloc[0]
        >>> ci.EnergyPriceForecast().get_prices_shortterm_archive(modified_date=cat_dates['modifiedDate'], category_id=cat_dates['categoryId'], month=12, delivery_region="Singapore")
        """
        path = "prices-short-term-archive"
        filter_params: list[str] = []

        if year_gt != None:
            filter_params.append(f"year > {year_gt}")
        if year_gte != None:
            filter_params.append(f"year >= {year_gte}")
        if year_lt != None:
            filter_params.append(f"year < {year_lt}")
        if year_lte != None:
            filter_params.append(f"year <= {year_lte}")

        filter_params.append(list_to_filter("year", year))
        filter_params.append(list_to_filter("month", month))
        filter_params.append(list_to_filter("priceSymbol", symbol))
        filter_params.append(list_to_filter("unitName", units))
        filter_params.append(list_to_filter("categoryName", category))
        filter_params.append(list_to_filter("groupName", group))
        filter_params.append(list_to_filter("currencySymbol", currency))
        filter_params.append(list_to_filter("commodityName", commodity))
        filter_params.append(list_to_filter("deliveryRegionName", delivery_region))
        filter_params.append(list_to_filter("sectorName", sector))

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {
            "filter": filter_exp,
            "page": page,
            "pageSize": page_size,
            "PriceCategoryCode": category_id,
            "ModifiedDate": modified_date,
        }
        return get_data(
            path=f"{self._path}{path}",
            params=params,
            paginate=paginate,
            raw=raw,
            paginate_fn=self._paginate,
            df_fn=self._to_df,
        )

    def get_prices_longterm_archive(
        self,
        *,
        modified_date: date,
        category_id: int,
        year: Optional[Union[list[int], "Series[int]", int]] = None,
        year_gt: Optional[int] = None,
        year_gte: Optional[int] = None,
        year_lt: Optional[int] = None,
        year_lte: Optional[int] = None,
        symbol: Optional[Union[list[str], "Series[str]", str]] = None,
        units: Optional[Union[list[str], "Series[str]", str]] = None,
        currency: Optional[Union[list[str], "Series[str]", str]] = None,
        category: Optional[Union[list[str], "Series[str]", str]] = None,
        group: Optional[Union[list[str], "Series[str]", str]] = None,
        sector: Optional[Union[list[str], "Series[str]", str]] = None,
        commodity: Optional[Union[list[str], "Series[str]", str]] = None,
        delivery_region: Optional[Union[list[str], "Series[str]", str]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[DataFrame, Response]:
        """
        Archived long-term price forecasts.

        see `get_archive_dates()` for applicable modified dates/categories.

        Parameters
        ----------
        modified_date : date
            filter on ModifiedDate to see the nearest prior forecast
        category_id : int
            filter on PriceCategoryCode
        year : Optional[Union[int, list[int], Series[int]]], optional
            filter by ``year = x``, by default None
        year_gt : Optional[int], optional
            filter by ``year > x``, by default None
        year_gte : Optional[int], optional
            filter by ``year >= x``, by default None
        year_lt : Optional[int], optional
            filter by ``year < x``, by default None
        year_lte : Optional[int], optional
            filter by ``year <= x``, by default None
        symbol : Optional[Union[list[str], Series[str], str]], optional
            filter by priceSymbol, by default None
        units : Optional[Union[list[str], Series[str], str]], optional
            filter by unitName, by default None
        currency : Optional[Union[list[str], Series[str], str]], optional
            filter by currencySymbol, by default None
        category : Optional[Union[list[str], Series[str], str]], optional
            filter by categoryName, by default None
        group : Optional[Union[list[str], Series[str], str]], optional
            filter by groupName, by default None
        sector : Optional[Union[list[str], Series[str], str]], optional
            filter by sectorName, by default None
        commodity : Optional[Union[list[str], Series[str], str]], optional
            filter by commodityName, by default None
        delivery_region : Optional[Union[list[str], Series[str], str]], optional
            filter by deliveryRegionName, by default None
        filter_exp : Optional[str], optional
            pass-thru ``$filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to select a certain page number, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object

        Examples
        --------
        **Simple**
        >>> from datetime import date
        >>> ci.EnergyPriceForecast().get_prices_longterm_archive(modified_date=date(2020, 11, 1), category_id=1, year=2022)

        **Using Lists**
        >>> from datetime import date
        >>> ci.EnergyPriceForecast().get_prices_longterm_archive(modified_date=date(2020, 11, 1) category_id=1, year=2022, group=["Refinery Margins", "U.S. West Coast"])

        **Using Series**
        >>> cat_dates = ci.EnergyPriceForecast().get_archive_dates(term_type="long", category="Oil")
        >>> cat_dates = cat_dates.iloc[0]
        >>> ci.EnergyPriceForecast().get_prices_longterm_archive(modified_date=cat_dates['modifiedDate'], category_id=cat_dates['categoryId'], delivery_region="Singapore")
        """
        path = "prices-long-term-archive"
        filter_params: list[str] = []

        if year_gt != None:
            filter_params.append(f"year > {year_gt}")
        if year_gte != None:
            filter_params.append(f"year >= {year_gte}")
        if year_lt != None:
            filter_params.append(f"year < {year_lt}")
        if year_lte != None:
            filter_params.append(f"year <= {year_lte}")

        filter_params.append(list_to_filter("year", year))
        filter_params.append(list_to_filter("priceSymbol", symbol))
        filter_params.append(list_to_filter("unitName", units))
        filter_params.append(list_to_filter("categoryName", category))
        filter_params.append(list_to_filter("groupName", group))
        filter_params.append(list_to_filter("currencySymbol", currency))
        filter_params.append(list_to_filter("commodityName", commodity))
        filter_params.append(list_to_filter("deliveryRegionName", delivery_region))
        filter_params.append(list_to_filter("sectorName", sector))

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {
            "filter": filter_exp,
            "page": page,
            "pageSize": page_size,
            "PriceCategoryCode": category_id,
            "ModifiedDate": modified_date,
        }
        return get_data(
            path=f"{self._path}{path}",
            params=params,
            paginate=paginate,
            raw=raw,
            paginate_fn=self._paginate,
            df_fn=self._to_df,
        )

    def get_reference_data(
        self, type: RefTypes, raw: bool = False
    ) -> Union[Response, DataFrame]:
        """
        Fetch reference data for the Energy Price Forecast dataset.

        Parameters
        ----------
        type : RefTypes
            filter by type
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
                DataFrame of the ``response.json()``
            Response
                Raw ``requests.Response`` object

        Examples
        --------
        **Simple**
        >>> ci.EnergyPriceForecast().get_reference_data(type=ci.EnergyPriceForecast.RefTypes.Sectors)
        """
        path = type.value

        params = {"pageSize": 1000, "page": 1}

        return get_data(
            path=f"{self._path}{path}",
            params=params,
            paginate=True,
            paginate_fn=self._paginate,
            raw=raw,
        )

    def get_archive_dates(
        self,
        term_type: Literal["short", "long"],
        category: Optional[Union[list[str], "Series[str]", str]],
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Get archive dates to use as a `modified_date` when calling the `short-term-archive` or `long-term-archive` methods.

        Parameters
        ----------
        term_type : Literal[short, long]
            whether to get short-term or long-term forecast dates.
        category : Optional[Union[list[str], Series[str], str]]
            filter on the categoryName
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
              DataFrame of the ``response.json()``
            Response
              Raw ``requests.Response`` object

        Examples
        --------
        **Simple**
        >>> ci.EnergyPriceForecast().get_archive_dates(term_type="short", category="Oil")
        """
        path = f"{term_type}-term-archive-dates"

        filter_params: list[str] = []

        filter_params.append(list_to_filter("categoryName", category))

        filter_params = [fp for fp in filter_params if fp != ""]

        params = {
            "filter": filter_params,
            "pageSize": 1000,
            "page": 1,
            "groupBy": "modifiedDate, categoryId, categoryName",
        }
        return get_data(
            path=f"{self._path}{path}",
            params=params,
            raw=raw,
            paginate=True,
            paginate_fn=self._paginate,
            df_fn=self._to_df,
        )
